#!/usr/bin/env python
from .backwards_search import backwards_search
import pickle

def mapper_main(args):

    ref_output = "{}.silly".format(args.reference)
    with open(ref_output, 'rb') as ref_fh:
        count_lookup, rank, burrows_wheeler, sa_index = pickle.load(ref_fh)

    total_length = len(sa_index)
    with open(args.reads) as reads_fh:
        for i, line in enumerate(reads_fh):
            if i % 4 == 0:
                read_id = line.strip()[1:]
            if i % 4 == 1:
                line = line.strip()
                s, e = backwards_search(line, count_lookup, rank, total_length)
                if s <= e:
                    print("{},{}".format(read_id, sa_index[s]))



def backwards_search(p, cl, rank, total_length):
    """Perform backwards search using pattern p, count lookup cl and rank rank

    reference http://alexbowe.com/fm-index/"""

    start = 0
    end = total_length - 1
    
    for i in range(len(p)-1,-1,-1):
        if end < start:
            break
        
        char = p[i]
        count_for_char = cl[char]
        
        start = count_for_char + rank.rank(start-1, char)
        end = count_for_char + rank.rank(end, char) - 1

    return start, end



import pickle
from .count_lookup import count_lookup
from .rank_lookup import Rank
from .burrows_wheeler import burrows_wheeler

def index_main(args):
    with open(args.reference, 'r') as ref_fh:
        reference_seq = ""
        for i, line in enumerate(ref_fh):
            if line.startswith('>'):
                if i != 0:
                    raise ValueError("Silly you, you expect sillymap to accept more than one reference sequence?")
                reference_id = line[1:].strip()
            else:
                reference_seq += line.strip()
    
    bw, sa_index = burrows_wheeler(reference_seq)
    cl = count_lookup(bw)
    rank = Rank()
    rank.add_text(bw)

    ref_output = "{}.silly".format(args.reference)
    with open(ref_output, "wb") as pickle_fh:
        pickle.dump((cl, rank, bw, sa_index), pickle_fh) 



from collections import Counter

def count_lookup(text):
    """Returns a lookup table for c, returning the 
    number of characthers in text lexically smaller than c.
    """
    char_count = Counter(text)
    lookup = {}

    current_count = 0
    for c in sorted(list(set(text))):
        lookup[c] = current_count
        current_count += char_count[c]

    return lookup




def burrows_wheeler(text):
    """Calculates the burrows wheeler transform of <text>.

    returns the burrows wheeler string and the suffix array indices
    The text is assumed to not contain the character $"""

    text += "$"
    all_permutations = []
    for i in range(len(text)):
        all_permutations.append((text[i:] + text[:i],i))

    all_permutations.sort()
    bw_l = [] # burrows wheeler as list
    sa_i = [] # suffix array indices

    for w,j in all_permutations:
        bw_l.append(w[-1])
        sa_i.append(j)

    return "".join(bw_l), sa_i



from collections import defaultdict
import warnings

class Rank():
    """Rank table corresponding to http://alexbowe.com/fm-index/ except 0-indexed.

    The rank(i,c) is the number of occurencies of c within text[0:(i+1)]

    public methods:
    add_text(text) -- initialize the rank for text.
    rank(i,c) -- returns rank of character c in intercal [0:i] 

    ex:

    rank = Rank()
    rank.add_text("ipssm$pissii")
    rank.rank(-1,i) = 0
    rank.rank("ipssm$pissii")[0,i] = 1
    rank.rank("ipssm$pissii")[11,s] = 4
    """

    def __init__(self):
        self.rank_dict = {}
        self.constructed = False

    def add_text(self, text):
        char_count = defaultdict(int)
        for i, c in enumerate(text):
            char_count[c] += 1
            for char, count in char_count.items():
                if not i in self.rank_dict:
                    self.rank_dict[i] = {}

                self.rank_dict[i][char] = count

        self.constructed = True

    def rank(self, i, c):
        if not self.constructed:
            warnings.warn(("Fetched rank before text was added, "
                           "this will always return 0")) 
        if not i in self.rank_dict:
            return 0
        if not c in self.rank_dict[i]:
            return 0
        return self.rank_dict[i][c]



from setuptools import setup, find_packages

setup(name='sillymap',
        version='0.1.0',
        description='A silly implementation of a short read mapper for the course DD3436.',
        url='http://github.com/alneberg/sillymap',
        author='Johannes Alneberg',
        author_email='alneberg@kth.se',
        license='MIT',
        packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
        zip_safe=False,
        scripts=['bin/sillymap']
   )
